/*
 *	Simple Mail Queue Manager -- Front End
 *
 *	(c) 2000 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>

unsigned int pid;
time_t timestamp;

void
genname(char *where, char *prefix)
{
  static unsigned int cnt = 0;
  unsigned int hash = 259309*pid + 123122323*(unsigned int)timestamp + 1000003*cnt;
  cnt++;
  sprintf(where, "%s/%08x", prefix, hash);
}

int
main(int argc, char **argv)
{
  char tmpname[32], newname[32], buf[4096];
  int destfd, maxiter, i, uid;
  pid_t daemon_pid;

  setgid(getegid());
  setuid(geteuid());
  openlog("enqueue", LOG_PID, LOG_MAIL);

  if (argc > 2)
    {
      syslog(LOG_ERR, "Invalid arguments");
      return 1;
    }
  if (argc == 2)
    {
      unsigned char *c = argv[1];
      if (*c == '@')
	c++;
      while (*c && (isalnum(*c) || *c == '_' || *c == '-'))
	c++;
      if (*c || (unsigned)(c - (unsigned char *) argv[1]) > sizeof(buf) - 8)
	{
	  syslog(LOG_ERR, "Insecure usage");
	  return 1;
	}
      sprintf(buf, "%s\n", argv[1]);
    }
  else
    strcpy(buf, "\n");

  if (chdir(HOMEDIR) < 0)
    {
      syslog(LOG_ERR, "Cannot change to " HOMEDIR);
      return 1;
    }
  pid = getpid();
  timestamp = time(NULL);

  maxiter = 1000;
  do
    genname(tmpname, "tmp");
  while ((destfd = open(tmpname, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0 && errno == EEXIST && --maxiter);
  if (destfd < 0)
    {
      syslog(LOG_ERR, "Cannot create temporary file: %m");
      return 1;
    }

  i = strlen(buf);
  do
    {
      if (write(destfd, buf, i) != i)
	{
	  syslog(LOG_ERR, "Error while writing temporary file: %m");
	  return 1;	/* The daemon will clean it up later */
	}
    }
  while ((i = read(0, buf, sizeof(buf))) > 0);
  if (i < 0)
    {
      syslog(LOG_ERR, "Error reading input: %m");
      return 1;
    }
  close(destfd);

  maxiter = 100;
  do
    genname(newname, "queue");
  while ((i = link(tmpname, newname)) < 0 && errno == EEXIST && --maxiter);
  if (i < 0)
    {
      syslog(LOG_ERR, "Cannot move temporary file to queue: %m");
      return 1;
    }
  if (unlink(tmpname) < 0)
    {
      syslog(LOG_ERR, "Cannot unlink temporary file: %m");
      return 1;
    }

  buf[256] = 0;
  if ((destfd = open("usherd.pid", O_RDONLY)) >= 0 &&
      read(destfd, buf, 256) > 0 &&
      sscanf(buf, "%d", &daemon_pid) == 1)
    kill(daemon_pid, SIGUSR1);

  return 0;
}
